#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include "basic.h"
#include "treiber.h"

#define CAS(x,y,z) __sync_bool_compare_and_swap(x,y,z)

static int dbg = 0 ;

typedef stack_node_t * tagged_t ;
/* Nodes */

/* Nodes are allocated on block boundaries, so as to use
   the BSZ-1 lower bits for tag */


#define BSZ 8
#define BLOCK (1U << BSZ)
#define MASK (BLOCK-1)

static void *align(void *p) {
  uintptr_t x = (uintptr_t)p ;
  x += BLOCK-1 ;
  x /= BLOCK ;
  x *= BLOCK ;
  return (void *)x ;
}

static stack_node_t *getptr(tagged_t p) {
  uintptr_t x = (uintptr_t)p ;
  x &=  ~((uintptr_t)MASK) ;
  return (stack_node_t *)x ;
}


static unsigned gettag(tagged_t p) {
  uintptr_t x = (uintptr_t)p ;
  return x & MASK ;
}

static tagged_t pack(stack_node_t *p, unsigned tag) {
  uintptr_t x = (uintptr_t)p ;
  x &=  ~((uintptr_t)MASK) ;
  //  if (tag >= BLOCK) fprintf(stderr,"Warning tag wraparound\n") ;
  x |= (tag & MASK) ;
  return (tagged_t)x ;
}

static tagged_t alloc_stack_node(void *val) {
  void *p = malloc_check(sizeof(stack_node_t)+BLOCK) ;
  stack_node_t *r = align(p) ;
  r->val = val ; r->next = NULL ;
  r->block = p ;
  return r ;
}

static void free_stack_node(stack_node_t *p) { free(p->block) ; }

stack_t *alloc_stack(int nprocs) {
  stack_t *r =  malloc_check(sizeof(*r)) ;
  r->top = NULL ;
  return r ;
}

void free_stack(stack_t *p) { free(p) ; }


void push(stack_t *p, void *v) {
  stack_node_t *q = alloc_stack_node(v) ;
  tagged_t old,q_packed ;
  do {
    SYNC() ;
    old = p->top ;
    stack_node_t *realold = getptr(old) ;
    unsigned tag = gettag(old) ;
    SYNC() ;
    q->next = realold ;                        
    SYNC() ;
    q_packed = pack(q,tag) ;
  } while (!CAS(&p->top,old,q_packed)) ; // Do push
  SYNC() ;
  if (dbg) fprintf(stderr,"PUSH: %p\n",q) ;
}

void *pop(stack_t *p,int id) {
  for ( ; ; ) {
    SYNC() ;
    tagged_t old = p->top ; // Old top node.
    SYNC() ;
    stack_node_t *realold = getptr(old) ;
    unsigned tag = gettag(old) ;
    if (realold == NULL) {
      return NULL ;
    } else {
      stack_node_t *realnext = realold->next ;
      SYNC() ;
      tagged_t next = pack(realnext,tag+1) ;
      if (CAS(&p->top,old,next)) { // Do pop
        SYNC() ;
        if (dbg) fprintf(stderr,"POP: %p\n",old) ;
        void *r = realold->val ;         // Clean up and return
        free_stack_node(realold) ;
        return r ;
      }
    }
  }
}
